import re, sys, yaml
from markdown_it import MarkdownIt
from mdit_py_plugins.front_matter import front_matter_plugin

from ..config import Config
from ..modes import get_mode
from ..utils import printer, expand_path, default

from .context import ScriptContext
from .section import ScriptSection

class ScriptRuntime:

  def __init__(self, text, data='', params: dict = None):
    params = default(params, {})
    self.md = MarkdownIt().use(front_matter_plugin)

    (self.meta, self.sections) = self.parse(text)
    config = Config()

    Mode = get_mode(self.meta.get('mode', 'synth-chat'))
    self.mode = Mode(profile=config.load_profile(Mode.default_profile_name))
    self.context = ScriptContext(self.mode, data=data, params=params)

    # I'll remove this hard coding soon
    self.mode.seed = 'You must only respond with exactly what was requested. Don\'t start your response with text like "Here is..."'
    self.mode.max_response_tokens = 750

  def run(self):
    for (i, section) in enumerate(self.sections):
      section.bind(self.mode, self.context)
      output = section.expand()

      printer.print_markdown(f'## [{i}] Input')
      printer.print_markdown(output)
      printer.print_markdown(f'## [{i}] Output')
      for (i, data) in enumerate(self.mode.ask(output)):
        if i == 0:
          data = data.lstrip()
        print(data, end='')
        sys.stdout.flush()
      print('\n')
      printer.print_markdown('-------------------------------')
    return self.mode.conversation[-1]['text']

  def parse(self, text):
    tokens = self.md.parse(text)

    meta = {}
    if tokens[0].type == 'front_matter':
      meta = yaml.safe_load(tokens[0].content)
      text = '\n'.join(text.split('\n')[tokens[0].map[1]:])

    chunks = re.split(r'^\*{3,}\s*$', text, flags=re.MULTILINE)
    sections = [
      self.create_section(chunk)
      for chunk in chunks
      if chunk.strip() != ''
    ]

    return (meta, sections)

  def create_section(self, text):
    return ScriptSection(self.md, text)

  @classmethod
  def exec(cls, name='', path='', code='', data='', params: dict = None):
    params = default(params, {})
    if name:
      config = Config()
      code = config.get_script(name)
    elif path:
      with open(expand_path(path), 'r') as file:
        code = file.read()
    elif code:
      pass

    script = cls(code, data=data, params=params)

    return script.run()



import typer, sys
from typing import Optional, List

from .config import Config
from .repl import REPL
from .repl.search import fuzzy_search_input
from .modes import list_modes
from .utils import printer
from .script import ScriptRuntime

config = Config()

# I'd like to have completion but I find the completion flags in --help distracting
app = typer.Typer(
  add_completion=False,
  pretty_exceptions_enable=False,
  pretty_exceptions_short=False,
  help="Try 'lmtk repl'",
)

@app.command()
def repl(
    thread: Optional[str] = typer.Argument(None),
    mode: str = typer.Option('synth-chat', '--mode', '-m'),
    profile: str =typer.Option(None, '--profile', '-p'),
):
  """
  """
  if thread == '@':
    thread = fuzzy_search_input('thread = @', config.threads().list()) or None
  elif thread == '-':
    thread = config.threads().list()[0]
  elif thread == 'i':
    thread = fuzzy_search_input('thread = @', config.threads().list()) or None
    mode = fuzzy_search_input('mode = ', list_modes()) or None
  if not thread:
    thread = config.threads().make_name()

  REPL(
    thread_name=thread,
    mode_name=mode,
    profile_name=profile,
  ).run()

@app.command()
def script(
    name: str,
    params: Optional[List[str]] = typer.Argument(None)
):
  params_dict = {}
  for p in params:
    parts = p.split('=')
    if len(parts) == 2:
      params_dict[parts[0]] = parts[1]

  if name.endswith('.md'):
    path = name
    name = None
  else:
    path = None

  printer.toggle_syntax_guessing(False)
  result = ScriptRuntime.exec(
    name=name,
    path=path,
    data=sys.stdin.read() if not sys.stdin.isatty() else '',
    params=params_dict,
  )
  if result:
    printer.print_markdown(f'# Result\n{result}')

@app.command()
def notebook():
  print('notebook')

@app.command()
def modes():
  for mode in list_modes():
    print(mode)

@app.command()
def threads():
  for thread in config.threads().list():
    print(thread)

def apply_aliases():
  if len(sys.argv) < 2:
    sys.argv += [ '--help' ]
    return
  cmd = sys.argv[1]

  if cmd[0] == '@':
    sys.argv[1] = 'repl'
    sys.argv.insert(2, cmd)
  elif cmd == '-':
    sys.argv[1] = 'repl'
    sys.argv.insert(2, cmd)
  elif cmd[0] == '.':
    sys.argv[1] = 'script'
    sys.argv.insert(2, cmd[1:])
  elif cmd[0] == '_':
    sys.argv[1] = 'notebook'
    sys.argv.insert(2, cmd[1:])
  elif cmd == 'help':
    sys.argv.pop(1)
    sys.argv += [ '--help' ]

def run():
  apply_aliases()
  app()

if __name__ == '__main__':
  run()



from .misc import peek, clear_screen, copy_to_clipboard, make_iter, expand_path, render_code_display, DotDict, default, make_list
from .printer import printer
from .loader import Loader
from .editor import open_in_editor
from .server import SimpleServer



import os, sys, tiktoken
from ...utils import printer

def check_api_key():
  if not os.environ.get('OPENAI_API_KEY'):
    printer.print("[red]Please set the [blue]OPENAI_API_KEY[/blue] environment variable.[/red]\n\nIf you don't have one you can generate one here https://beta.openai.com/account/api-keys")
    sys.exit(1)

tokenizer = None
def get_tokenizer():
  global tokenizer
  tokenizer = tokenizer or tiktoken.get_encoding('gpt2')
  return tokenizer

def count_tokens(text):
  return len(get_tokenizer().encode(text))



import requests, subprocess, os
from bs4 import BeautifulSoup
from prompt_toolkit.shortcuts import prompt
from ..utils import expand_path

def get_web(url):
  if not url.startswith('http'):
    url = f'https://{url}'
  html = requests.get(url).text
  soup = BeautifulSoup(html, 'html.parser')
  return soup.get_text().strip()

def get_file(file_path):
  with open(expand_path(file_path), 'r') as file:
    return file.read()

def show_code(file_name):
  file_path = expand_path(file_name)
  if not os.path.exists(file_path):
    raise FileNotFoundError(f'File "{file_name}" does not exist')

  file_content = get_file(file_path)
  if not file_content:
    return ''

  extension = file_name.split('.')[-1]
  return f"""
`{ file_path }`
```{extension}
{ file_content }
```
""".strip()

def show_web(url):
  return f"""
`{ url }`
```
{ get_web(url) }
```
""".strip()

def shell(cmd):
  return subprocess.getoutput(cmd)

def ask(question=''):
  if question:
    return prompt(question + '\n')
  else:
    return prompt()



"""
This is a port of a small subset of OpenAI's embedding utils. The goal is to
decrease the number of dependencies. Many of the dependencies in the original
are for creating plots.

https://github.com/openai/openai-python/blob/main/openai/embeddings_utils.py
"""

from typing import List
from tenacity import retry, stop_after_attempt, wait_random_exponential

import numpy as np
import openai

default_engine = 'text-embedding-ada-002'

@retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(6))
def get_embedding(text: str, engine=default_engine) -> List[float]:

    # replace newlines, which can negatively affect performance.
    text = text.replace("\n", " ")

    return openai.Embedding.create(input=[text], engine=engine)["data"][0]["embedding"]

@retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(6))
def get_embeddings(
    list_of_text: List[str], engine=default_engine
) -> List[List[float]]:
    assert len(list_of_text) <= 2048, "The batch size should not be larger than 2048."

    # replace newlines, which can negatively affect performance.
    list_of_text = [text.replace("\n", " ") for text in list_of_text]

    data = openai.Embedding.create(input=list_of_text, engine=engine).data
    data = sorted(data, key=lambda x: x["index"])  # maintain the same order as input.
    return [d["embedding"] for d in data]

def cosine_similarity(a, b):
    """
    According to the OpenAI docs, embeddings are now already normalized, so
    I've removed the normalization step.
    """
    return np.dot(a, b)
    # return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))

def distances_from_embeddings(
    query_embedding: List[float],
    embeddings: List[List[float]],
) -> List[List]:
    """Return the distances between a query embedding and a list of embeddings."""
    return [
        cosine_similarity(query_embedding, embedding)
        for embedding in embeddings
    ]

def indices_of_nearest_neighbors_from_distances(distances) -> np.ndarray:
    """Return a list of indices of nearest neighbors from a list of distances."""
    return np.argsort(distances)



class LmtkError(Exception):
  pass

class LmtkApiError(LmtkError):

  def __init__(self, error):
    self.error = error
    self.message = getattr(error, 'message', '') or getattr(error, '_message', '') or ''

  def __repr__(self):
    return repr(self.error)



from .base_mode import BaseMode, register_mode
from ..llms.open_ai import GPT3

@register_mode('raw-gpt')
class RawGPTMode(BaseMode):

  title = 'RawGPT'

  def load(self, state):
    self.llm = GPT3()
    self.model = 'text-davinci-003'
    self.temperature = .5

  def respond(self, query):
    completion = self.llm.complete(
      query,
      model=self.model,
      temperature=float(self.temperature),
      stream=True,
      # soft_stops=[ '1003', '3373', '15211', '11900' ],
    )

    prefix = self.parse_prefix(query)
    for (i, data) in enumerate(completion):
      if i == 0 and prefix:
        yield prefix
      yield data

  def parse_prefix(self, query):
    parts = query.split(':>')
    return parts[-1].strip() if len(parts) > 1 else ''

  def inspect(self):
    return ''.join(
      [ m['text'] for m in self.conversation[-2:] ]
    )

  def variables(self):
    return [
      'temperature',
    ]



import sys
from ..utils import DotDict, default, expand_path

from .helpers import get_web, get_file, shell, ask, show_web, show_code

class ScriptContext:

  def __init__(self, mode, data='', params: dict = None):
    params = default(params, {})

    self.mode = mode
    self.output_buffer = ''

    self.build_vars(data, params)

  def build_vars(self, data='', params: dict = None):
    params = default(params, {})
    self.live_vars = {
      'data': data,
      'params': DotDict(params),
      'args': sys.argv[2:],
      'echo': lambda s: self.echo(s),
      'get_web': get_web,
      'get_file': get_file,
      'get_path': expand_path,
      'shell': shell,
      'ask': ask,
      'show_web': show_web,
      'show_code': show_code
    }

  def set_var(self, key, val):
    self.live_vars[key] = val

  def prepare(self):
    self.clear_output()
    answers = [ m['text'] for m in self.mode.conversation if m['source'] == 'server']
    self.live_vars['answers'] = answers
    self.live_vars['__'] = answers if len(answers) > 0 else ''

  def get_output(self):
    return self.output_buffer

  def clear_output(self):
    self.output_buffer = ''

  def get_vars(self):
    return self.live_vars

  def echo(self, s):
    self.output_buffer += s

  def eval(self, code, inline=False):
    self.prepare()
    if inline:
      result = eval(code, globals(), self.get_vars())
    else:
      exec(code, globals(), self.get_vars())
      result = self.get_output()
    return str(result)



import os
from ..utils import expand_path, default

class Folders:

  def __init__(self, config_path=None):
    self.config_path = expand_path(config_path)
    self.folders = {}

    self.add('config', self.config_path)
    self.add('builtin', expand_path(__file__, '..', '..', 'builtin'))

  def add(self, folder_name, folder_path=None):
    folder_path = default(folder_path, expand_path(self.config_path, folder_name))

    if not os.path.exists(folder_path):
      os.makedirs(folder_path)
    self.folders[folder_name] = folder_path

    return folder_path

  def get_path(self, folder_name):
    if folder_name in self.folders:
      return self.folders[folder_name]
    else:
      return None

  def get_files(self, folder_name):
    folder_path = self.get_path(folder_name)
    if not folder_path:
      return None
    return os.listdir(folder_path)

  def get_file_path(self, folder_name, file_name, optional_extensions=[]):
    if type(file_name) is list:
      file_name = os.path.join(*file_name)

    file_path = expand_path(self.get_path(folder_name), file_name)

    for ext in optional_extensions:
      file_with_ext = file_path + ext
      if ext and os.path.isfile(file_with_ext):
        return file_with_ext

    return file_path

  def load_file(self, folder_name, file_name, optional_extensions=[]):
    folder_path = self.get_path(folder_name)
    if not folder_path:
      return None
    file_path = self.get_file_path(folder_name, file_name, optional_extension)
    with open(file_path, 'r') as file:
      return file.read()

  def create_file(self, folder_name, file_name, file_content):
    folder_path = self.get_path(folder_name)
    if not folder_path:
      return None
    file_path = expand_path(folder_path, file_name)
    with open(file_path, 'w') as file:
      file.write(file_content)
    return file_path



import uuid
from ..utils import make_iter, SimpleServer, printer, default

from ..config.profile import Profile

mode_registry = {}

def register_mode(name):
  def decorator(cls):
    mode_registry[name] = cls
    return cls
  return decorator

def get_mode(name):
  return mode_registry[name]

def list_modes(show_hidden=False):
  modes = []
  for name, mode in mode_registry.items():
    if name != 'base' and mode.visible or show_hidden:
      modes += [ name ]
  modes.sort()
  return modes

@register_mode('base')
class BaseMode:

  title = ''
  visible = True
  loader_latency=1.5
  default_profile_name = None

  seed = ''

  web_server_config = None
  web_server = None
  __has_logged = False

  def __init__(self, state: dict = None, profile=None):
    self.state = default(state, {})
    self.profile = profile if profile != None else Profile()

    self.conversation = self.state.get('conversation', [])
    self.seed = self.state.get('seed', '')
    self.buffers = {}
    self.active = True
    self.load(self.state)

    if self.web_server_config != None:
      self.__serve()

  def ask(self, query):
    client_message = self.__add_message(text=query, source='client')
    server_message = self.__build_message(source='server')

    response = make_iter(self.respond(query))
    for data in response:
      server_message['text'] += data
      yield data

    self.__add_message(message=server_message)

  def save_state(self):
    state = {
      'conversation': self.conversation,
      'seed': self.seed,
    }
    sub_state = self.save() or {}
    state.update(sub_state)
    return state

  def stop(self):
    if self.active:
      self.unload()
    if self.web_server != None:
      self.web_server.stop()
    self.active = False

  def rollback_n(self, n=1):
    for i in range(n):
      self.rollback()
      if len(self.conversation) > 0:
        self.conversation.pop()

  def __build_message(self, text='', source=''):
    message = {
      'id': str(uuid.uuid4()),
      'text': text,
      'source': source,
    }
    return message

  def __add_message(self, message=None, text='', source=''):
    if not message:
      message = self.__build_message(text=text, source=source)
    self.conversation += [ message ]
    return message

  def __serve(self):
    self.web_server = SimpleServer(
      lambda request, path: self.request_handler(request, path),
      host=self.web_server_config.get('host', 'localhost'),
      port=self.web_server_config.get('port', 8080),
    )
    success = self.web_server.start()

    if BaseMode.__has_logged:
      return
    BaseMode.__has_logged = True

    if success:
      printer.print(f'[bold]Notice:[/bold] Sandbox available at [bold]{self.web_server.host}:{self.web_server.port}[/bold] \n')
    else:
      printer.print(f'[bold]Warning:[/bold] Failed to start sandbox web server. Port {self.web_server.port} is in use.\n')

  # To use ipdb set:
  #   export PYTHONBREAKPOINT=IPython.terminal.debugger.set_trace
  def debug(self):
    breakpoint()

  ##############################
  # Can override starting here #
  ##############################

  def respond(self, query):
    yield ''

  def load(self, state):
    pass

  def unload(self):
    pass

  def save(self):
    return {}

  def get_title(self):
    return self.title

  def get_seed(self):
    return self.seed

  def set_seed(self, seed):
    self.seed = seed

  def rollback(self):
    pass

  def stats(self):
    return ''

  def inspect(self):
    return ''

  def variables(self):
    return []

  def get_buffer(self, name=''):
    return (name or 'default', self.buffers.get(name, ''), '.md')

  def set_buffer(self, name, value):
    self.buffers[name] = value

  def request_handler(self, request, path):
    return ''



import threading, socket
from http.server import HTTPServer, BaseHTTPRequestHandler

class RequestHandler(BaseHTTPRequestHandler):

  def do_GET(self):
    """Handle a GET request by calling the handler function and returning the response."""
    self.send_response(200)
    self.end_headers()
    result = self.server.handler(self.request, self.path) or ''
    self.wfile.write(result.encode())

  def log_request(self, *args, **kwargs):
    pass
  def log_request_line(self, *args, **kwargs):
    pass
  # def log_error(self, *args, **kwargs):
  #   pass
  def log_message(self, *args, **kwargs):
    pass

class SimpleServer(threading.Thread):
  def __init__(self, handler, host='localhost', port=8080):
    """
    Initialize the threaded HTTP server with the specified host, port, and handler function.
    The handler function should take in a request object and return the response string.
    """
    self.handler = handler
    self.host = host
    self.port = port
    self.running = False

    super().__init__(daemon=True)

  def start(self):
    if self.is_port_used(self.port):
      return False

    self.httpd = HTTPServer((self.host, self.port), RequestHandler)
    self.httpd.handler = self.handler
    super().start()

    return True

  def run(self):
    """Start the HTTP server in a separate thread."""
    self.running = True
    self.httpd.serve_forever()

  def stop(self):
    """Stop the HTTP server."""
    if self.running:
      self.httpd.shutdown()
      self.httpd.socket.close()

  def is_port_used(self, port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
      return s.connect_ex(('localhost', port)) == 0



import openai
from ...errors import LmtkApiError
from .utils import check_api_key, count_tokens, get_tokenizer

class GPT3:

  def __init__(self):
    check_api_key()
    get_tokenizer() # preload tokenizer

  def count_tokens(self, text):
    return count_tokens(text)

  def complete(self, *args, **kwargs):
    if kwargs.get('stream'):
      return self.get_response_async(*args, **kwargs)
    else:
      return self.get_response_sync(*args, **kwargs)

  def get_response_async(self, *args, **kwargs):
    try:
      response = self.get_response(*args, **kwargs)
      for data in response:
        yield data.choices[0].text
    except openai.OpenAIError as error:
      raise LmtkApiError(error)

  def get_response_sync(self, *args, **kwargs):
    try:
      response = self.get_response(*args, **kwargs)
      return response.choices[0].text
    except openai.OpenAIError as error:
      raise LmtkApiError(error)

  def get_response(self,
      prompt,
      max_length=1000,
      temperature=0.7,
      model="text-davinci-003",
      stops=None,
      soft_stops=[],
      stream=False,
    ):

    logit_bias = {}
    for s in soft_stops:
      logit_bias[s] = -100

    return openai.Completion.create(
      engine=model,
      prompt=prompt,
      max_tokens=max_length,
      temperature=temperature,
      top_p=1,
      frequency_penalty=0,
      presence_penalty=0,
      stop=stops,
      logit_bias=logit_bias,
      stream=stream,
      # logit_bias={
      #   '2949': 1, # Make "No" a more common token
      #   '44651': 1, # Make "Invalid" a more common token
      # },
    )


  def get_model_max_tokens(self, model):
    if model == 'text-davinci-003':
      return 4000
    elif model == 'code-davinci-002':
      return 8000
    else:
      return 2000



import uuid, re, os, json
from datetime import datetime
from .message import Message
from ..modes import get_mode

class Thread:

  id = ''
  name = ''
  mode_name = ''
  mode_state = {}
  profile_name = ''
  seed = ''
  # timestamp = None

  def __init__(self, name, config):
    self.set_name(name)
    self.config = config
    self.mode = None

    if not self.load():
      self.reset()

  def load_mode(self):
    if self.mode:
      self.mode.stop()
    self.mode = get_mode(self.mode_name)(
      state=self.mode_state,
      profile=self.get_profile(),
    )
    self.mode.set_seed(self.seed)
    return self.mode

  def stop_mode(self):
    if self.mode:
      self.mode.stop()
      self.mode = None

  def set_mode(self, mode_name, state=None):
    self.mode_name = mode_name
    self.mode_state = state or self.mode_state

  def get_profile(self):
    profile_name = self.profile_name or get_mode(self.mode_name).default_profile_name
    return self.config.load_profile(profile_name)

  def set_profile(self, profile_name):
    self.profile_name = profile_name

  def get_file_path(self):
    file_name = f'{self.escape_name(self.name)}.json'
    return self.config.folders.get_file_path('threads', file_name)

  def to_data(self):
    return {
      'id': self.id,
      'head_id': self.head_id,
      'name': self.name,
      'all_messages': { msg_id: msg.to_data() for (msg_id, msg) in self.all_messages.items() },
      'mode_name': self.mode_name,
      'mode_state': self.mode_state,
      'profile_name': self.profile_name,
      'seed': self.seed,
      # 'timestamp': self.timestamp,
    }

  def load_data(self, data):
    self.id = data.get('id')
    self.head_id = data.get('head_id')
    self.name = data.get('name')
    self.all_messages = {
      msg_id: Message().load_data(msg_data)
      for (msg_id, msg_data) in data.get('all_messages', {}).items()
    }
    self.mode_name = data.get('mode_name')
    self.mode_state = data.get('mode_state')
    self.profile_name = data.get('profile_name')
    self.seed = data.get('seed')
    # self.timestamp = data.get('timestamp')
    return self

  def save(self, stop=False):
    if self.mode:
      self.mode_state = self.mode.save_state()
      self.seed = self.mode.get_seed()

    file_path = self.get_file_path()
    with open(file_path, 'w') as thread_file:
      json.dump(self.to_data(), thread_file, indent=2)

    if stop:
      self.stop_mode()

  def load(self):
    file_path = self.get_file_path()
    if not os.path.isfile(file_path):
      return False

    with open(file_path, 'r') as thread_file:
      self.load_data(json.load(thread_file))
    return True

  def set_name(self, name):
    self.name = self.normalize_name(name)

  def reset(self, preserve_profile=False, preserve_seed=False):
    self.id = self.id or str(uuid.uuid4())
    self.head_id = None
    self.all_messages = {}
    self.mode_name = self.mode_name or ''
    self.mode_state = {}
    self.profile_name = self.profile_name if preserve_profile else ''
    self.seed = self.seed if preserve_seed else ''
    # self.timestamp = datetime.now()

  def get_messages(self, head_id=None):
    head_id = head_id or self.head_id
    messages = []
    while head_id:
      msg = self.all_messages[head_id]
      messages.insert(0, msg)
      head_id = msg.parent_id
    return messages

  def add_message(self, source, text, stats=''):
    message = Message(source, text, stats=stats, parent_id=self.head_id)
    self.all_messages[message.id] = message
    self.head_id = message.id
    return message

  def rollback_n(self, n=1):
    new_head_id = self.head_id
    for i in range(n):
      msg = self.all_messages.get(new_head_id)
      if not msg:
        break
      new_head_id = msg.parent_id
    self.head_id = new_head_id

  @classmethod
  def normalize_name(cls, thread_name):
    thread_name = thread_name.replace('@', '').replace('_', '-').strip()
    return re.sub(r'[^-a-zA-Z0-9\.]+', '_', thread_name)

  @classmethod
  def escape_name(cls, thread_name):
    return cls.normalize_name(thread_name).replace('-', '_')



from prompt_toolkit import PromptSession
from prompt_toolkit.completion import FuzzyWordCompleter
from prompt_toolkit.shortcuts import prompt

def fuzzy_search_input(prefix, options, erase=True):
  session = PromptSession(erase_when_done=erase)
  completer = FuzzyWordCompleter(options)
  return session.prompt(
    prefix,
    completer=completer,
  )



# INSTALLATION:
#   Copy or symlink this file into `$LMTK_CONFIG_PATH/plugins/`
#   By default, this will be `~/.config/lmtk/plugins/`

from lmtk.modes import BaseMode, register_mode
from lmtk.llms.open_ai import GPT3

@register_mode('bruh')
class BruhMode(BaseMode):

  title = 'Bruh'

  def load(self, state):
    self.llm = GPT3()

  def respond(self, query):
    prompt = f'Rephrase the following text to bruh-speak:\n{query}\n\nBruh:\n'
    return self.llm.complete(
      prompt,
      stream=True
    )



import re

class ScriptSection:

  tag_regex = re.compile(r'{{(.*?)}}')

  def __init__(self, md, text):
    self.md = md
    self.source_text = text
    self.expanded_text = None

  def bind(self, mode, context):
    self.mode = mode
    self.context = context

  def expand(self):
    self.expanded_text = self.source_text

    tokens = self.md.parse(self.source_text)
    changes = []
    for token in tokens:
      changes += self.expand_token(token)

    if len(changes) == 0:
      return self.display()
    changes.sort(key=lambda x: x[0][0])

    spans = [ v[0] for v in changes ]
    lines = self.source_text.split('\n')
    new_lines = []

    new_lines += lines[:spans[0][0]]
    for (i, change) in enumerate(changes):
      if i != 0:
        new_lines += lines[spans[i-1][1] + 1 : spans[i][0]]
      new_lines += [ change[1] ]
    new_lines += lines[spans[-1][1]+1:]

    self.expanded_text = '\n'.join(new_lines)
    return self.display()

  def expand_token(self, token):
    if token.map == None or not token.content:
      return []

    text_lines = self.source_text.split('\n')[token.map[0]:token.map[1]]
    text = '\n'.join(text_lines)

    if token.type == 'fence' and (text_lines[0] == '```eval' or text_lines[1] == '#eval'):
      result = self.context.eval(token.content)
      return [ (token.map, result) ]
    elif re.search(self.tag_regex, text):
      result = re.sub(self.tag_regex, lambda m: self.context.eval(m.group(1), True), text)
      return [ (token.map, result) ]
    else:
      return []

  def display(self):
    return self.expanded_text



import html
from itertools import chain

# Only used if minify=True. Not sure if it's worth the extra deps
# from jsmin import jsmin
# from jsbeautifier import beautify

from .base_mode import BaseMode, register_mode
from ..llms.open_ai import GPT3
from ..utils import printer, render_code_display, DotDict

@register_mode('js-sandbox')
class JSSandboxMode(BaseMode):

  title = 'JS Sandbox'
  minify = False

  web_server_config = {
    'host': 'localhost',
    'port': 8081,
  }

  prompt_prefix = '```javascript\n'
  stops = [  '/*END*/', '```', '// END', '//END' ]
  has_logged = False

  default_profile_name = 'js-blank'

  def load(self, state):
    self.model = 'text-davinci-003'
    self.llm = GPT3()

    self.history = state.get('history', [])

    self.sandbox = self.profile.config
    self.inner_html = self.sandbox.get('inner_html', '')
    self.bio = self.sandbox.get('bio', 'You are an experienced software engineer')

    sandbox_starter = self.sandbox.get('starter_code', '')
    min_sandbox_starter = self.sandbox.get('min_starter_code', sandbox_starter)

    self.file_name = 'index.min.js' if self.minify else 'index.js'
    self.starter_code = min_sandbox_starter if self.minify else sandbox_starter
    self.code = state.get('code', self.starter_code)

  def save(self):
    return {
      'code': self.code,
      'history': self.history,
    }

  def respond(self, query):
    self.history += [ { 'text': query, 'type': 'client' } ]
    code = ''
    results = self.complete(self.get_prompt(query))

    if self.minify:
      code = ''.join(list(results))
      yield self.prompt_prefix
      yield self.beautify_code(code)
    else:
      yield self.prompt_prefix
      for data in results:
        code += data
        yield data

    yield '\n```'

    self.history += [ { 'text': self.code, 'type': 'server' } ]
    self.code = code

  def complete(self, text):
    return self.llm.complete(
      text,
      stream=True,
      max_length=self.llm.get_model_max_tokens(self.model) - self.llm.count_tokens(text),
      model=self.model,
      stops=self.stops,
    )

  def request_handler(self, request, path):
    if path == '/':
      return render_code_display(
        code=self.beautify_code(self.code),
        frame='/sandbox',
      )
    elif path == '/sandbox':
      return self.render_html(
        code=self.code,
        inner_html=self.inner_html,
        style=f'body {{ margin: 0px; }}\n{self.sandbox.get("style", "")}',
      )

  def inspect(self):
    return self.get_prompt('{ instruction }')

  def rollback(self):
    if len(self.history) > 0:
      msg = self.history.pop()
      if msg['type'] == 'server':
        self.code = msg['text']
    else:
      self.code = self.starter_code

  def get_buffer(self, name):
    return ('code', self.code, '.js')

  def set_buffer(self, name, value):
    self.code = value

  def stats(self):
    return f'( tokens={len(self.get_prompt(""))} )'

  @property
  def loader_latency(self):
    return .1 if self.minify else 1.5

  def minify_code(self, code):
    return jsmin(code) if self.minify else code.rstrip()

  def beautify_code(self, code):
    return beautify(code) if self.minify else code

  def get_prompt(self, instruction=''):
    return f"""
web/client/index.html
```html
{self.render_html(inner_html=self.inner_html)}
```
web/client/index.js
```javascript
{self.minify_code(self.code)}
/*END*/
```

{self.bio}. Make the following modifications to `index.js`:
{instruction}

web/client/{self.file_name}
{ self.prompt_prefix }"""

  def render_html(self, code=None, inner_html='', style=None):
    if code == None:
      script_tag = '<script src="index.js"></script>'
    else:
      script_tag = f'<script>{code}</script>'

    if style == None:
      style_tag = ''
    else:
      style_tag = f'<style>{style}</style>';

    return f"""
<html>
  <head>{style_tag}</head>
  <body>{inner_html}{script_tag}</body>
</html>
""".strip()



from .base_mode import BaseMode, register_mode
from ..llms.open_ai import GPT3

@register_mode('raw-codex')
class RawCodexMode(BaseMode):

  title = 'Codex'

  def load(self, state):
    self.llm = GPT3()

  def respond(self, query):
    yield '```\n'
    for data in self.llm.complete(
        text,
        stream=True,
        model='code-davinci-002',
        temperature=0.1,
    ):
      yield data
    yield '\n```'



import os, re, pyperclip, html
from itertools import chain
from collections.abc import Iterable

def peek(gen):
  first = next(gen)
  return ( chain([ first ], gen), first )

def make_iter(x):
  if isinstance(x, str) or not isinstance(x, Iterable):
    return iter([ x ])
  else:
    return x

def make_list(x):
  return list(make_iter(x))

def expand_path(*args):
  return os.path.abspath(os.path.expanduser(os.path.join(*args)))

# Yeah, sorry not sorry. This is needed because of a
# confluence of my 3 least favorite parts of Python:
#   1. Shared mutable default parameters
#   2. Falsey {}, [], etc. For some random class `C`, is
#      `C()` falsey? Who knows... it is user defined.
#   3. Ugly and non-linear "ternary" statements
# I mean... wtf. Aside from some arguably wacky variable
# scoping and the mess that is package management, that's
# pretty much the complete list of things I strongly
# dislike about Python.
#
# I know Python doesn't support macros so this can't
# short-circuit, but whatever. At least `default` isn't
# a reserved parameter. So rejoice, dear reader.
def default(val, fallback):
  return val if val != None else fallback

class DotDict(dict):
  __getattr__ = dict.__getitem__
  __setattr__ = dict.__setitem__
  __delattr__ = dict.__delitem__

  def __init__(self, source: dict = None):
    for key, value in default(source, {}).items():
      self[key] = value

class set_env:

  def __init__(self, name, value):
    self.name = name
    self.value = value

  def __enter__(self):
    self.old_value = os.environ.get(self.name)
    os.environ[self.name] = self.value

  def __exit__(self, *args):
    if self.old_value == None:
      del os.environ[self.name]
    else:
      os.environ[self.name] = self.old_value

def clear_screen():
  os.system('cls' if os.name == 'nt' else 'clear')

def copy_to_clipboard(text):
  try:
    pyperclip.copy(text)
    return True
  except Exception as e:
    return False

# This doesn't really belong here, but it's here for now
def render_code_display(code='', frame='', language='javascript'):
  formatted_code = html.escape(code)
  return f"""
<html style="height: 100%">
  <head>
    <style>
      html {{
        height: 100%;
      }}
      body {{
        display: flex;
        flex-direction: row;
        background: #444654;
      }}
      .row {{
        flex: 1;
        margin: 10px;
        margin-top: 15px;
      }}
      iframe {{
        border: 0px;
        background: #ededed;
        box-shadow: 0px 0px 20px #000;
      }}
      pre {{
        display: inline-block;
        margin-top: 0px;
        text-align: left;
      }}
      code {{
        height: 774px;
        width: 700px;
        box-shadow: 0px 0px 20px #000;
      }}
      #fullscreen {{
        position: fixed;
        bottom: 1rem;
        right: 1rem;
        color: rgb(217,217,227);
        background-color: rgba(52,53,65);
        border-color: rgba(86,88,105);
        font-family: Helvetica;
        padding: 10px;
        box-shadow: 0px 0px 3px #000;
        text-decoration: none;
      }}
    </style>
    <link  rel="stylesheet" href="https://fonts.googleapis.com/css?family=Roboto">
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/styles/monokai.min.css">
    <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/highlight.min.js"></script>
  </head>
  <body>
    <div class="row" style="text-align: right">
      <iframe width="700" height="800" src="{frame}"></iframe>
    </div>
    <div class="row" style="text-align: left">
      <pre><code class="language-{language}" id="code">{formatted_code}</code></pre>
      <script>
        hljs.highlightAll();
      </script>
    </div>
    <a href="{frame}" target="_blank" id="fullscreen">Fullscreen</a>
  </body>
</html>
"""



from dataclasses import dataclass
import openai
import numpy as np # openai already depends on this

from .utils import check_api_key

vec_dim = 1536

class GPTEmbeddingManager:

  def __init__(self, model='text-embedding-ada-002'):
    check_api_key()

    self.model = model

    self.sources = []
    self.embeddings = []

  def calculate_embeddings(self, texts):
    if len(texts) == 0:
      return []

    # replace newlines, which can negatively affect performance.
    texts = [ text.replace('\n', ' ') for text in texts ]

    result = openai.Embedding.create(
      input=texts,
      model=self.model,
    )
    return [ r['embedding'] for r in result['data'] ]

  def calculate_embedding(self, text):
    return self.calculate_embeddings([ text ])[0]

  def add(self, *args, **kwargs):
    if isinstance(args[0], str):
      self.add_one(*args, **kwargs)
    else:
      self.add_multiple(*args, **kwargs)

  def add_one(self, text, meta=None, embedding=None):
    meta = {} if meta is None else meta

    self.sources.append({
      'index': len(self.embeddings),
      'text': text,
      'meta': meta,
    })

    if embedding == None:
      self.embeddings.append(self.calculate_embedding(text))
    else:
      self.embeddings.append(embedding)

  def add_multiple(self, inputs):
    for (i, input) in enumerate(inputs):
      if not isinstance(input, tuple):
        inputs[i] = (input, {})

    texts = [ text for (text, _) in inputs ]
    embeddings = self.calculate_embeddings(texts)

    for (i, embedding) in enumerate(embeddings):
      self.add(texts[i], meta=inputs[i][1], embedding=embedding)

  def search(self, text, top_n=5):
    if len(self.embeddings) == 0:
      return []

    embedding = self.calculate_embedding(text)
    similarities = np.matmul(self.embeddings, embedding)

    # similarities = np.matmul(np.asmatrix(manager.embeddings), embedding)
    # similarities = np.asarray(similarities)[0]

    top_indices = similarities.argsort()[::-1][:top_n]
    return [
      SearchResult(similarity=similarities[i], **self.sources[i])
      for i in top_indices
    ]

@dataclass
class SearchResult:
  similarity : float
  index: int
  text: str
  meta: dict



import os, yaml
from ..utils import default, DotDict

# This class should always do dependency injection and support
# optional constructor args so it can be a useful shim without
# triggering a cascade of imports.
class Profile:

  def __init__(self, name=None, folders=None):
    self.name = name or ''
    self.folders = folders

    config = self.load()

    if config == None:
      self.name = ''
      self.config = DotDict({})
      self.empty = True
      self.mode = None
    else:
      self.config = DotDict(config)
      self.empty = False
      self.mode = self.config.get('mode', None)

  def load(self):
    if not self.name or not self.folders:
      return None

    file_path = self.find_file(self.name)
    if file_path is None:
      return None

    with open(file_path, 'r') as f:
      return yaml.safe_load(f)

  def find_file(self, name):
    file_path = self.folders.get_file_path('builtin', [ 'profiles', name ], [ '.yaml', '.yml' ])
    if os.path.exists(file_path):
      return file_path

    file_path = self.folders.get_file_path('profiles', name, [ '.yaml', '.yml' ])
    if os.path.exists(file_path):
      return file_path

  def exists(self):
    return len(self.config) > 0



import time
from ipykernel.kernelbase import Kernel

from ..modes import get_mode
from ..config import Config
Mode = get_mode('synth-chat')

class LmtkKernel(Kernel):
    implementation = 'lmtk'
    implementation_version = '1.0'
    language = 'no-op'
    language_version = '0.1'
    language_info = {
        'name': 'Any text',
        'mimetype': 'text/markdown',
        'file_extension': '.md',
    }
    banner = 'lmtk - Kernel for attaching to lmtk'

    def __init__(self, *args, **kwargs):
        self.my_log('\n---- START ----')
        self.lmtk_config = Config()
        super().__init__(*args, **kwargs)

    def my_log(self, txt):
        log_file = open('/tmp/kernel-log.txt', 'w')
        log_file.write(txt)
        log_file.close()

    def do_execute(self,
        code,
        silent,
        store_history=True,
        user_expressions=None,
        allow_stdin=False,
    ):
        if not silent:
            self.my_log(f'\nInput: {code}\nOutput:\n')
            self.mode = Mode(profile=self.lmtk_config.load_profile(Mode.default_profile_name))
            text = ''

            display_id = str(time.time())
            self.send_response(self.iopub_socket, 'display_data', {
                'metadata': {},
                'data': {
                    'text/plain': 'Loading...',
                },
                'transient': {
                    'display_id': display_id,
                },
            })

            for (i, data) in enumerate(self.mode.ask(code)):
                if i == 0:
                    data = data.lstrip()
                self.my_log(data)
                text += data
                self.send_response(self.iopub_socket, 'update_display_data', {
                    'metadata': {},
                    'data': {
                        'text/plain': text + '█',
                    },
                    'transient': {
                        'display_id': display_id,
                    },
                })

            self.send_response(self.iopub_socket, 'update_display_data', {
                'metadata': {},
                'data': {
                    'text/markdown': text,
                },
                'transient': {
                    'display_id': display_id,
                },
            })
        return {
          'status': 'ok',
          'execution_count': self.execution_count,
          'payload': [],
          'user_expressions': {},
        }

if __name__ == '__main__':
    from ipykernel.kernelapp import IPKernelApp
    IPKernelApp.launch_instance(kernel_class=LmtkKernel)



import os, json, re, importlib.util
from ..utils import printer, expand_path
from ..threads import ThreadManager
from .folders import Folders
from .profile import Profile

class Config:

  default_config_path = '~/.config/lmtk'

  def __init__(self, config_path=None):
    self.config_path = config_path or os.getenv('LMTK_CONFIG_PATH', self.default_config_path)
    self.folders = Folders(self.config_path)

    config_folders = [
      'threads',
      'plugins',
      'scripts',
      'profiles',
    ]
    for folder in config_folders:
      self.folders.add(folder)

    self.config_file_path = self.folders.get_file_path(
      'config',
      'config.json'
    )
    self.prompt_history_path = self.folders.get_file_path(
      'config',
      'prompt_history'
    )

    self.reload()

  def save(self):
      with open(self.config_file_path, 'w') as config_file:
        json.dump(self.config or {}, config_file, indent=2)

  def reload(self):
    self.config = {}

    if not os.path.exists(self.config_file_path):
      self.save()

    with open(self.config_file_path, 'r') as config_file:
      self.config = json.load(config_file)

    return self.config

  def threads(self):
    return ThreadManager(self)

  def load_profile(self, profile_name):
    return Profile(profile_name, self.folders)

  def get_script(self, script_name):
    file_path = self.folders.get_file_path('scripts', script_name, [ '.md' ])
    if not file_path:
      return None
    with open(file_path, 'r') as file:
      return file.read()

  def load_plugins(self):
    files = self.folders.get_files('plugins')
    for file in files:
      file_path = self.folders.get_file_path('plugins', file)
      if not file_path.endswith('.py'):
        continue
      try:
        spec = importlib.util.spec_from_file_location('plugins', file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
      except Exception as e:
        printer.exception(e)
        continue
